#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gtk/gtk.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <gdk/gdkx.h>
#include <malloc.h>
#include <string.h>
#define WNCK_I_KNOW_THIS_IS_UNSTABLE
#include <libwnck/libwnck.h>
#include <libwnck/class-group.h>

#define ICON_SIZE 72

GdkCursor * moveCursor;
GSList * tiles = NULL;
gboolean needs_sync=FALSE;
GSList * pendingActivations=NULL;
int lastEventTime=0;

#define MAX_HORZ 256
#define MAX_VERT 256

Atom DockAppAtom;
Atom RestackAtom;
Atom ActiveWinAtom;
GtkTooltips * tips = NULL;
int width,height;
int activatedCount=0;

struct _BDockTile;
typedef struct _BDockTile BDockTile;

#define TILECOMMON \
    BDockTile * anchor; \
    int x; \
    int y;

typedef struct _BDockSwallow
{
    TILECOMMON;
    GdkNativeWindow id;
} BDockSwallow;

typedef struct _BDockApp
{
    TILECOMMON;
    WnckClassGroup *    class_group;
    gchar *             class_name;
    gchar *             command;
    gboolean            sticky;
    GdkPixbuf *         inactive_img;
} BDockApp;

typedef struct _BDockAnchor
{
    gboolean    contains[MAX_HORZ][MAX_VERT];
    int         width;
    int         height;
    int         x;
    int         y;
} BDockAnchor;

typedef union _BDockData
{
    BDockApp        app;
    BDockAnchor     anchor;
    BDockSwallow    swallow;
} BDockData;

typedef enum _BDockClass
{
    BDOCK_CLASS_APP,
    BDOCK_CLASS_ANCHOR,
    BDOCK_CLASS_SWALLOW,
    BDOCK_CLASS_COUNT
} BDockClass;

typedef enum _BDockAction
{
    BDOCK_ACTION_NONE,
    BDOCK_ACTION_MOVE,
    BDOCK_ACTION_END_MOVE,
    BDOCK_ACTION_LAUNCH,
    BDOCK_ACTION_SWITCH_TO,
    BDOCK_ACTION_SHOW_ALL,
    BDOCK_ACTION_HIDE_ALL,
    BDOCK_ACTION_TOGGLE_HIDDEN,
    BDOCK_ACTION_MENU,
    BDOCK_ACTION_COLLAPSE,
    BDOCK_ACTION_EXPAND,
    BDOCK_ACTION_COUNT
} BDockAction;

typedef enum _BDockButton
{
    BDOCK_BUTTON1_DOWN,
    BDOCK_BUTTON2_DOWN,
    BDOCK_BUTTON3_DOWN,
    BDOCK_BUTTON1_UP,
    BDOCK_BUTTON2_UP,
    BDOCK_BUTTON3_UP,
    BDOCK_BUTTON1_DBL,
    BDOCK_BUTTON2_DBL,
    BDOCK_BUTTON3_DBL,
    BDOCK_BUTTON_COUNT
} BDockButton;

typedef enum _BDockAnchorState
{
    BDOCK_ANCHOR_STATE_EXPANDED,
    BDOCK_ANCHOR_STATE_COLLAPSED,
    BDOCK_ANCHOR_STATE_COUNT
} BDockAnchorState;

typedef enum _BDockSwallowState
{
    BDOCK_SWALLOW_STATE_RUNNING,
    BDOCK_SWALLOW_STATE_STOPPED,
    BDOCK_SWALLOW_STATE_COUNT
} BDockSwallowState;

typedef enum _BDockAppState
{
    BDOCK_APP_STATE_RUNNING_INACTIVE,
    BDOCK_APP_STATE_RUNNING_ACTIVE,
    BDOCK_APP_STATE_NOT_RUNNING,
    BDOCK_APP_STATE_COUNT
} BDockAppState;

typedef enum _BDockModifier
{
    BDOCK_MODIFIER_NONE,
    BDOCK_MODIFIER_COUNT
} BDockModifier;

typedef enum _BDockGenericState
{
    S0,S1,S2,S3,S4,S5
} BDockGenericState;

typedef union _BDockState
{
    BDockGenericState   generic;
    BDockAppState       app;
    BDockAnchorState    anchor;
    BDockSwallowState   swallow;
} BDockState;

struct _BDockTile
{
    GtkWindow *     win;
    
    GtkWidget *     child;
    GtkWidget *     fixed;
    GtkWidget *     menu;
    GtkWidget *     bg;

    GdkPixmap *     bgimg;
    GdkPixmap *     child_bgimg;

    int             child_off_x;
    int             child_off_y;
    int             child_width;
    int             child_height;

    BDockClass      class;
    BDockData       data;
    BDockState      state;
    int             w;
    int             h;
    gboolean        dragging;
    gboolean        beginning_drag;
    int             drag_button;
    gdouble         dragX;
    gdouble         dragY;
};

typedef enum _FindTileMode
{
    FIND_TILE_BY_CLASS_GROUP,
    FIND_TILE_BY_CLASS_NAME,
} FindTileMode;

typedef struct _FindDockTile
{
    FindTileMode        mode;
    BDockClass          class;
    WnckClassGroup *    class_group;
    const gchar *       class_name;
    gboolean            found;
    BDockTile *         tile;
} FindDockTile;
BDockAction SwallowActions[BDOCK_MODIFIER_COUNT][BDOCK_BUTTON_COUNT][BDOCK_SWALLOW_STATE_COUNT]=
{
    {   //NO MODIFIER//
    //   RUNNING                    STOPPED
        {BDOCK_ACTION_MOVE,         BDOCK_ACTION_MOVE       }, //1DOWN
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2DOWN
        {BDOCK_ACTION_MENU,         BDOCK_ACTION_MENU       }, //3DOWN
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_LAUNCH     }, //1UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //3UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //1DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //3DBL
    },
};
BDockAction AnchorActions[BDOCK_MODIFIER_COUNT][BDOCK_BUTTON_COUNT][BDOCK_ANCHOR_STATE_COUNT]=
{
    {   //NO MODIFIER//
    //   EXPANDED                   COLLAPSED
        {BDOCK_ACTION_MOVE,         BDOCK_ACTION_MOVE       }, //1DOWN
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2DOWN
        {BDOCK_ACTION_MENU,         BDOCK_ACTION_MENU       }, //3DOWN
        {BDOCK_ACTION_COLLAPSE,     BDOCK_ACTION_EXPAND     }, //1UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //3UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //1DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //2DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE       }, //3DBL
    },
};
BDockAction AppActions[BDOCK_MODIFIER_COUNT][BDOCK_BUTTON_COUNT][BDOCK_APP_STATE_COUNT]=
{
    {   //NO MODIFIER//
    //   RUNNING INACTIVE           RUNNING ACTIVE              NOT RUNNING
        {BDOCK_ACTION_MOVE,         BDOCK_ACTION_MOVE,          BDOCK_ACTION_MOVE       }, // 1DOWN
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 2DOWN
        {BDOCK_ACTION_MENU,         BDOCK_ACTION_MENU,          BDOCK_ACTION_MENU       }, // 3DOWN
        {BDOCK_ACTION_SWITCH_TO,    BDOCK_ACTION_HIDE_ALL,      BDOCK_ACTION_LAUNCH     }, // 1UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 2UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 3UP
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 1DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 2DBL
        {BDOCK_ACTION_NONE,         BDOCK_ACTION_NONE,          BDOCK_ACTION_NONE       }, // 3DBL
    },
};

#define NEW(b,a) b * a = malloc(sizeof(b)); memset((a),0,sizeof(b))
void find_tile_func(BDockTile * tile, FindDockTile * fdt)
{
    if (!fdt->found && fdt->class==tile->class && ( 
            (fdt->mode==FIND_TILE_BY_CLASS_GROUP && 
                    tile->data.app.class_group == fdt->class_group) ||
            (fdt->mode==FIND_TILE_BY_CLASS_NAME && 
                    !strcmp(tile->data.app.class_name,fdt->class_name)) ||
            FALSE
            ))
    {
        fdt->found=TRUE;
        fdt->tile=tile;
    }
}
BDockTile * find_tile_by_class_group(WnckClassGroup * group)
{
    FindDockTile fdt;
    fdt.found=FALSE;
    fdt.mode=FIND_TILE_BY_CLASS_GROUP;
    fdt.class=BDOCK_CLASS_APP;
    fdt.class_group=group;
    g_slist_foreach(tiles,(GFunc)find_tile_func,&fdt);
    if (fdt.found)
        return fdt.tile;
    else
        return NULL;
}
BDockTile * find_tile_by_class_name(const gchar * name)
{
    FindDockTile fdt;
    fdt.found=FALSE;
    fdt.class=BDOCK_CLASS_APP;
    fdt.mode=FIND_TILE_BY_CLASS_NAME;
    fdt.class_name=name;
    g_slist_foreach(tiles,(GFunc)find_tile_func,&fdt);
    if (fdt.found)
        return fdt.tile;
    else
        return NULL;
}

void place_at_anchor(BDockTile * ba)
{
    if (ba->class==BDOCK_CLASS_APP || ba->class==BDOCK_CLASS_SWALLOW)
    {
        int mx=ba->data.app.x*ICON_SIZE+ba->data.app.anchor->data.anchor.x;
        int my=ba->data.app.y*ICON_SIZE+ba->data.app.anchor->data.anchor.y;
        gtk_window_move(ba->win,mx,my);
    }
}

gboolean sync_timeout(gpointer p)
{
    if (needs_sync)
    {
        needs_sync=FALSE;
        g_slist_foreach(tiles,(GFunc)place_at_anchor,NULL);
    }
    return TRUE;
}
void do_set_icon_geometry(WnckWindow * w, BDockTile * ba)
{
    int x,y;
    x=ba->data.app.x*ICON_SIZE+ba->data.app.anchor->data.anchor.x;
    y=ba->data.app.y*ICON_SIZE+ba->data.app.anchor->data.anchor.y;
    wnck_window_set_icon_geometry(w,x,y,ICON_SIZE,ICON_SIZE);
}
void item_moved(int dx, int dy, BDockTile * ba)
{
    int mx,my;
    int tx,ty;
    gtk_window_get_position(ba->win,&tx,&ty);
    tx+=dx;
    ty+=dy;
    switch(ba->class)
    {
        case BDOCK_CLASS_APP:
        case BDOCK_CLASS_SWALLOW:
            mx=ba->data.app.x*ICON_SIZE+ba->data.app.anchor->data.anchor.x;
            my=ba->data.app.y*ICON_SIZE+ba->data.app.anchor->data.anchor.y;
            if (!ba->dragging && (tx != mx || ty != my))
            {
                int px,py;
                //figure the nearest square
                if (tx<ba->data.app.anchor->data.anchor.x)
                    px=0;
                else
                    px=(((tx-ba->data.app.anchor->data.anchor.x)+ICON_SIZE/2)/ICON_SIZE);
                if (ty<ba->data.app.anchor->data.anchor.y)
                    py=0;
                else
                    py=(((ty-ba->data.app.anchor->data.anchor.y)+ICON_SIZE/2)/ICON_SIZE);
                if (!ba->data.app.anchor->data.anchor.contains[px][py])
                {
                    ba->data.app.anchor->
                        data.anchor.contains[ba->data.app.x][ba->data.app.y]=FALSE;
                    ba->data.app.x=px;
                    ba->data.app.y=py;
                    ba->data.app.anchor->
                        data.anchor.contains[ba->data.app.x][ba->data.app.y]=TRUE;
                    mx=ba->data.app.x*ICON_SIZE+ba->data.app.anchor->data.anchor.x;
                    my=ba->data.app.y*ICON_SIZE+ba->data.app.anchor->data.anchor.y;
                    gtk_window_move(ba->win,mx,my);
                }
                else
                    gtk_window_move(ba->win,mx,my);
            }
            else
            {
                gtk_window_move(ba->win,tx,ty);
            }
            break;
        case BDOCK_CLASS_ANCHOR:
            if (ba->data.anchor.x != tx
                    || ba->data.anchor.y != ty)
            {
                mx=tx;
                my=ty;
                if (mx<0) mx=0;
                if (my<0) my=0;
                if (mx>(width-ICON_SIZE)) mx=width-ICON_SIZE;
                if (my>(height-ICON_SIZE)) my=height-ICON_SIZE;
                ba->data.anchor.x=mx;
                ba->data.anchor.y=my;
                gtk_window_move(ba->win,mx,my);
                needs_sync=TRUE;
            }
            break;
        default:
            g_warning("Invalid BDOCK_CLASS encountered");
    }
    if (ba->class==BDOCK_CLASS_APP && ba->data.app.class_group)
    {
        g_slist_foreach((GSList *)wnck_class_group_get_windows(
                    ba->data.app.class_group),
                (GFunc)do_set_icon_geometry,ba);
    }
}

void my_active_window_changed(WnckScreen * s, gpointer p)
{
    if (pendingActivations)
    {
        WnckWindow * w = wnck_screen_get_active_window(s);
        if (w==pendingActivations->data)
        {
            pendingActivations=g_slist_remove(pendingActivations,w);
        }
        if (pendingActivations)
        {
            wnck_window_activate(pendingActivations->data,lastEventTime);
        }
    }
}

void my_show_act_win(WnckWindow * w, gpointer p)
{
    if (!p && wnck_window_is_minimized(w))
        wnck_window_unminimize(w,lastEventTime);
    pendingActivations=g_slist_append(pendingActivations,w);
    activatedCount++;
}

void my_act_win(WnckWindow * w, WnckClassGroup * c)
{
    if (wnck_window_get_class_group(w)==c)
    {
        if (!(wnck_window_get_state(w) &
             (WNCK_WINDOW_STATE_HIDDEN |
              WNCK_WINDOW_STATE_MINIMIZED |
              WNCK_WINDOW_STATE_SHADED | // not sure about this one
              0
              )))
        {
            my_show_act_win(w,c);
        }
    }
}

void my_hide_win(WnckWindow * w, gpointer p)
{
    wnck_window_minimize(w);
}

void recalc_tile_state(BDockTile * ba)
{
    switch(ba->class)
    {
        case BDOCK_CLASS_APP:
            if (ba->data.app.class_group)
            {
                ba->state.app=
                    wnck_window_get_class_group(
                            wnck_screen_get_active_window(
                                wnck_screen_get_default()))==ba->data.app.class_group?
                    BDOCK_APP_STATE_RUNNING_ACTIVE:BDOCK_APP_STATE_RUNNING_INACTIVE;
            }
            else
                ba->state.app=BDOCK_APP_STATE_NOT_RUNNING;
            break;
        case BDOCK_CLASS_SWALLOW:
            ba->state.swallow=BDOCK_SWALLOW_STATE_RUNNING;
            break;
        default:
            break;
    }
}
void bdock_tile_show(BDockTile * bt)
{
    gtk_widget_show_all(GTK_WIDGET(bt->win));
}
void bdock_tile_hide(BDockTile * bt)
{
    gtk_widget_hide(GTK_WIDGET(bt->win));
}
void show_if_match(BDockTile * bt, BDockTile * anch)
{
    if (bt->class!=BDOCK_CLASS_ANCHOR && bt->data.app.anchor==anch)
        bdock_tile_show(bt);
}
void hide_if_match(BDockTile * bt, BDockTile * anch)
{
    if (bt->class!=BDOCK_CLASS_ANCHOR && bt->data.app.anchor==anch)
        bdock_tile_hide(bt);
}
void switch_to_item(GtkWidget * w, WnckWindow * win)
{
    wnck_window_activate(win,lastEventTime);
}
void append_class_item(WnckWindow * w, GtkMenuShell * menu)
{
    GtkWidget * mitem;
    mitem=gtk_image_menu_item_new_with_label(wnck_window_get_name(w));
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mitem),
            gtk_image_new_from_pixbuf(wnck_window_get_mini_icon(w)));
    g_signal_connect(mitem,"activate",G_CALLBACK(switch_to_item),w);
    gtk_widget_show(mitem);
    gtk_menu_shell_append(menu,mitem);
}

void bdock_tile_gen_menu(BDockTile * bt)
{
    GtkWidget * mitem;
    if (bt->menu)
        gtk_widget_destroy(bt->menu);
    bt->menu=gtk_menu_new();
    switch(bt->class)
    {
        case BDOCK_CLASS_APP:
            if (!bt->data.app.class_group)
                return;
            g_slist_foreach((GSList *)wnck_class_group_get_windows(bt->data.app.class_group),
                    (GFunc)append_class_item,bt->menu);
            break;
        case BDOCK_CLASS_ANCHOR:
            mitem=gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT,NULL);
            g_signal_connect(mitem,"activate",G_CALLBACK(gtk_main_quit),NULL);
            gtk_widget_show(mitem);
            gtk_menu_shell_append(GTK_MENU_SHELL(bt->menu),mitem);
            break;
        default:
            break;
    }
}

void show_menu(BDockTile * bt,GdkEvent * ev)
{
    if (!bt->menu)
    {
        bdock_tile_gen_menu(bt);
    }
    lastEventTime=((GdkEventButton *)ev)->time;
    if (ev->type==GDK_BUTTON_PRESS)
        gtk_menu_popup(GTK_MENU(bt->menu),
                NULL,NULL,NULL,NULL,((GdkEventButton *)ev)->button,
                gtk_get_current_event_time());
    else
        gtk_menu_popup(GTK_MENU(bt->menu),
                NULL,NULL,NULL,NULL,0,gtk_get_current_event_time());
}
void handle_tile_event(BDockAction act, BDockTile * ba, GdkEvent * ev)
{
    switch(act)
    {
        case BDOCK_ACTION_MENU:
            show_menu(ba,ev);
            break;
        case BDOCK_ACTION_MOVE:
            gtk_tooltips_disable(tips);
            ba->beginning_drag=TRUE;
            ba->drag_button=((GdkEventButton *)ev)->button;
            ba->dragX=((GdkEventButton *)ev)->x;
            ba->dragY=((GdkEventButton *)ev)->y;
            break;
        case BDOCK_ACTION_END_MOVE:
            gtk_tooltips_enable(tips);
            gdk_window_set_cursor(GTK_WIDGET(ba->win)->window,NULL);
            ba->dragging=FALSE;
            ba->beginning_drag=FALSE;
            item_moved(0,0,ba);
            sync_timeout(NULL);
            break;
        case BDOCK_ACTION_SWITCH_TO:
            {
                lastEventTime=((GdkEventButton *)ev)->time;
                g_slist_free(pendingActivations);
                pendingActivations=NULL;
                WnckScreen * scr = wnck_screen_get_default();
                activatedCount=0;
                GSList * wl = g_slist_copy(
                        (GSList *)wnck_screen_get_windows_stacked(scr));
                g_slist_foreach(wl,(GFunc)my_act_win,
                        ba->data.app.class_group);
                g_slist_free(wl);
                if (!activatedCount)
                {
                    wl=(GSList *)wnck_class_group_get_windows(ba->data.app.class_group);
                    g_slist_foreach(wl,(GFunc)my_show_act_win,NULL);
                }
                my_active_window_changed(scr,NULL);
            }
            break;
        case BDOCK_ACTION_SHOW_ALL:
            lastEventTime=((GdkEventButton *)ev)->time;
            g_slist_free(pendingActivations);
            pendingActivations=NULL;
            g_slist_foreach((GSList *)wnck_class_group_get_windows(ba->data.app.class_group),
                    (GFunc)my_show_act_win,NULL);
            break;
        case BDOCK_ACTION_HIDE_ALL:
            g_slist_free(pendingActivations);
            pendingActivations=NULL;
            g_slist_foreach((GSList *)wnck_class_group_get_windows(ba->data.app.class_group),
                    (GFunc)my_hide_win,NULL);
            break;
        case BDOCK_ACTION_LAUNCH:
            if (ba->data.app.command)
                g_spawn_command_line_async(ba->data.app.command,NULL);
            break;
        case BDOCK_ACTION_COLLAPSE:
            ba->state.anchor=BDOCK_ANCHOR_STATE_COLLAPSED;
            g_slist_foreach(tiles,(GFunc)hide_if_match,ba);
            break;
        case BDOCK_ACTION_EXPAND:
            ba->state.anchor=BDOCK_ANCHOR_STATE_EXPANDED;
            g_slist_foreach(tiles,(GFunc)show_if_match,ba);
            break;
        case BDOCK_ACTION_NONE:
            break;
        default:
            g_warning("Unhandled action %d",act);
            break;
    }
} 

void event_handler(GtkWidget * w, GdkEvent * ev, BDockTile * ba)
{
    int basebutt=0, butt;
    switch(ev->type)
    {
        case GDK_MOTION_NOTIFY:
            if (ba->dragging || ba->beginning_drag)
            {
                ba->dragging=TRUE;
                gdk_window_set_cursor(GTK_WIDGET(ba->win)->window,moveCursor);
                item_moved(((GdkEventMotion *)ev)->x-ba->dragX,
                        ((GdkEventMotion *)ev)->y-ba->dragY,ba);
            }
            break;
        case GDK_2BUTTON_PRESS:
            basebutt+=3;
        case GDK_BUTTON_RELEASE:
            if (ba->dragging)
            {
                if (ba->drag_button==((GdkEventButton *)ev)->button)
                {
                    handle_tile_event(BDOCK_ACTION_END_MOVE,ba,ev);
                    break;
                }
            }
            basebutt+=3;
        case GDK_BUTTON_PRESS:
            butt=((GdkEventButton*)ev)->button;
            if (butt<4 && butt>0)
            {
                BDockAction act;
                butt--;
                //hack this in for now TODO make event driven
                recalc_tile_state(ba);
                switch(ba->class) // hate to do this - work around C stupidity
                {
                    case BDOCK_CLASS_APP:
                        act=AppActions[0][butt+basebutt][ba->state.app];
                        break;
                    case BDOCK_CLASS_ANCHOR:
                        act=AnchorActions[0][butt+basebutt][ba->state.anchor];
                        break;
                    case BDOCK_CLASS_SWALLOW:
                        act=SwallowActions[0][butt+basebutt][ba->state.swallow];
                        break;
                    default:
                        act=BDOCK_ACTION_NONE;
                        g_warning("Invalid tile class");
                        break;
                }
                handle_tile_event(act,ba,ev);
            }
            break;
        default:
            break;
    }
}
void bdock_tile_init_placement(BDockTile * bt)
{
    int x,y;
    switch(bt->class)
    {
        case BDOCK_CLASS_ANCHOR:
            bt->data.anchor.width=width/ICON_SIZE;
            bt->data.anchor.height=height/ICON_SIZE;
            bt->data.anchor.contains[0][0]=TRUE;
            gtk_window_move(bt->win,0,0);
            break;
        case BDOCK_CLASS_APP:
        case BDOCK_CLASS_SWALLOW:
            if (bt->data.app.anchor->class!=BDOCK_CLASS_ANCHOR)
            {
                g_warning("Cannot place tiles when no anchor available");
                return;
            }
            //loop through the anchor's tile slots and place this tile
            for(y=0;y<bt->data.app.anchor->data.anchor.height;y++)
            {
                for(x=0;x<bt->data.app.anchor->data.anchor.width;x++)
                {
                    if (!bt->data.app.anchor->data.anchor.contains[x][y])
                    {
                        bt->data.app.anchor->data.anchor.contains[x][y]=TRUE;
                        bt->data.app.x=x;
                        bt->data.app.y=y;
                        item_moved(0,0,bt);
                        return;
                    }
                }
            }
            break;
        default:
            break;
    }
}

BDockTile * bdock_tile_new(BDockClass class, BDockTile * anchor)
{
    NEW(BDockTile,bt);

    //create and set up the tile's window
    GtkWindow * w;
    w = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
    gtk_container_set_border_width(GTK_CONTAINER(w),0);
    gtk_window_set_focus_on_map(w,FALSE);
    gtk_window_set_resizable(w,FALSE);
    gtk_window_set_decorated(w,FALSE);
    gtk_window_set_skip_pager_hint(w,TRUE);
    gtk_window_set_skip_taskbar_hint(w,TRUE);
    gtk_window_set_type_hint(w,GDK_WINDOW_TYPE_HINT_DOCK);
    gtk_window_stick(w);
    bt->win=w;
    GtkWidget * fi = gtk_fixed_new();
    gtk_container_set_border_width(GTK_CONTAINER(fi),0);
    bt->fixed=fi;
    gtk_container_add(GTK_CONTAINER(bt->win),fi);
    bt->w=ICON_SIZE;
    bt->h=ICON_SIZE;
    bt->bg=gtk_image_new();
    gtk_fixed_put(GTK_FIXED(bt->fixed),bt->bg,0,0);
    gtk_widget_set_size_request(fi,bt->w,bt->h);
    gtk_window_set_default_size(w,bt->w,bt->h);
    gtk_widget_set_size_request(GTK_WIDGET(w),bt->w,bt->h);
    bt->class=class;
    if (bt->class!=BDOCK_CLASS_ANCHOR)
        bt->data.app.anchor=anchor;
    bdock_tile_init_placement(bt);
    return bt;
}

void bdock_tile_update_child_bgimg(BDockTile * bt)
{
    if (bt->child_bgimg)
    {
        gdk_pixmap_unref(bt->child_bgimg);
        bt->child_bgimg=NULL;
    }
    gtk_widget_realize(bt->child);
    if (!bt->bgimg)
        return;
    if (GTK_WIDGET(bt->child)->window==GTK_WIDGET(bt->win)->window)
        return;
    bt->child_bgimg = gdk_pixmap_new(GTK_WIDGET(bt->child)->window,
            bt->child_width,bt->child_height,-1);
    GdkGC * gc = gdk_gc_new(GTK_WIDGET(bt->child)->window);
    gdk_draw_drawable(bt->child_bgimg,gc,
            bt->bgimg,bt->child_off_x,bt->child_off_y,0,0,
            bt->child_width,bt->child_height);
    gdk_gc_unref(gc);
    gtk_widget_set_app_paintable(bt->child,TRUE);
    gdk_window_set_back_pixmap(GTK_WIDGET(bt->child)->window,
            bt->child_bgimg,FALSE);
    if (GTK_WIDGET_FLAGS(GTK_WIDGET(bt->child)) & GTK_MAPPED)
        gtk_widget_queue_draw(bt->child);
}

void bdock_tile_set_bg_from_pixbuf(BDockTile * bt, GdkPixbuf * bgimg)
{
    if (bt->bgimg)
    {
        gdk_pixmap_unref(bt->bgimg);
        bt->bgimg=NULL;
    }
    gtk_widget_realize(GTK_WIDGET(bt->win));
    bt->bgimg = gdk_pixmap_new(GTK_WIDGET(bt->win)->window,
            bt->w,bt->h,-1);
    GdkGC * gc = gdk_gc_new(GTK_WIDGET(bt->win)->window);
    gdk_draw_pixbuf(bt->bgimg,gc,
            bgimg,0,0,0,0,bt->w,bt->h,
            GDK_RGB_DITHER_NONE,0,0);
    gdk_gc_unref(gc);
    gtk_image_clear(GTK_IMAGE(bt->bg));
    gtk_image_set_from_pixmap(GTK_IMAGE(bt->bg),bt->bgimg,NULL);
    if (bt->child)
        bdock_tile_update_child_bgimg(bt);
}

void bdock_tile_set_bg_from_file(BDockTile * bt, gchar * bgimg_path)
{
    //create the bg pixmap
    GdkPixbuf * p = gdk_pixbuf_new_from_file_at_size(bgimg_path,bt->w,bt->h,NULL);
    if (!p)
    {
        g_warning("Couldn't find bgimg %s",bgimg_path);
        return;
    }
    bdock_tile_set_bg_from_pixbuf(bt,p);
    gdk_pixbuf_unref(p);
}
void bdock_tile_draw_child_bg(GtkWidget * child, GdkEventExpose * ev, BDockTile * bt)
{
    if (!bt->child_bgimg)
        return;
    gdk_window_clear(GTK_WIDGET(child)->window);
    GdkGC * gc = gdk_gc_new(GTK_WIDGET(child)->window);
    gdk_draw_drawable(GTK_WIDGET(child)->window,gc,bt->child_bgimg,
            0,0,0,0,bt->child_width,bt->child_height);
    g_object_unref(gc);
}
void bdock_tile_set_child(BDockTile * bt, GtkWidget * child)
{
    //contain the child
    if (!bt->fixed)
        return;
    if (bt->child)
        gtk_container_remove(GTK_CONTAINER(bt->fixed),bt->child);
    bt->child=child;
    if (!child)
        return;
    gtk_fixed_put(GTK_FIXED(bt->fixed),child,0,0);
}
void bdock_tile_place_child(BDockTile * bt)
{
    if (!bt->child)
        return;
    int ofx,ofy;
    gint piw,pih;
    GtkRequisition req;
    gtk_widget_realize(bt->child);
    gtk_widget_size_request(bt->child,&req);
    piw=req.width;
    pih=req.height;
    if (piw==-1) piw=64;
    if (pih==-1) pih=64;
    ofx=(bt->w-piw)/2;
    ofy=(bt->h-pih)/2;
    if (bt->class==BDOCK_CLASS_SWALLOW)
    {
        g_signal_connect(bt->child,"expose-event",G_CALLBACK(bdock_tile_draw_child_bg),bt);
        ofx--;
        ofy--;
    }
    bt->child_off_x=ofx;
    bt->child_off_y=ofy;
    bt->child_width=piw;
    bt->child_height=pih;
    gtk_widget_set_size_request(bt->child,piw,pih);
    gtk_fixed_move(GTK_FIXED(bt->fixed),bt->child,ofx,ofy);
    bdock_tile_update_child_bgimg(bt);
}

void bdock_tile_register(BDockTile * bt)
{
    gtk_widget_add_events(GTK_WIDGET(bt->win),GDK_BUTTON_MOTION_MASK|
            GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
    g_signal_connect(bt->win,"event-after",G_CALLBACK(event_handler),bt);
    tiles=g_slist_append(tiles,bt);
}
void bdock_tile_destroy(BDockTile * bt)
{
    tiles=g_slist_remove_all(tiles,bt);
    if (bt->class!=BDOCK_CLASS_ANCHOR)
    {
        if (bt->data.app.anchor)
        {
            bt->data.app.anchor->data.anchor.contains
                [bt->data.app.x][bt->data.app.y]=FALSE;
        }
    }
    if (bt->bgimg)
        gdk_pixmap_unref(bt->bgimg);
    if (bt->child_bgimg)
        gdk_pixmap_unref(bt->child_bgimg);
    if (bt->win)
        gtk_widget_destroy(GTK_WIDGET(bt->win));
}
void set_dockapp_hint(WnckWindow * w)
{
    long int dat=1;
    gdk_error_trap_push();
    XChangeProperty(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            wnck_window_get_xid(w),DockAppAtom,XA_CARDINAL,32,PropModeReplace,
            (unsigned char *)&dat,1);
    XSync(GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),False);
    gdk_error_trap_pop();
}

gboolean has_dockapp_hint(WnckWindow * w)
{
    Atom atype;
    int afmt;
    unsigned long int nitems;
    unsigned long int naft;
    gboolean ret=FALSE;
    unsigned char * dat;
    gdk_error_trap_push();
    if (XGetWindowProperty(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            wnck_window_get_xid(w),DockAppAtom,0,1,False,
            XA_CARDINAL,&atype,&afmt,&nitems,&naft,&dat)==Success)
    {
        if (nitems==1 && ((long int *)dat)[0]==1)
        {
            ret=TRUE;
        }
        XFree(dat);
    }
    XSync(GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),False);
    gdk_error_trap_pop();
    return ret;
}


void checkit(WnckWindow * w, gboolean * val)
{
    XWMHints * h = XGetWMHints(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            wnck_window_get_xid(w));
    if (h && h->initial_state!=WithdrawnState)
    {
        if (!has_dockapp_hint(w))
        {
            /*WnckWindowType i;
            i = wnck_window_get_window_type(w);
            if (i!=WNCK_WINDOW_SPLASHSCREEN && i!=WNCK_WINDOW_DOCK
                    && i!=WNCK_WINDOW_DESKTOP)*/
            if (!(wnck_window_get_state(w) & 
                        (
                         WNCK_WINDOW_STATE_SKIP_TASKLIST |
                         WNCK_WINDOW_STATE_SKIP_PAGER)))
                *val=TRUE;
        }
    }
    if (h)
        XFree(h);
}
gboolean confirmApp(WnckClassGroup * a)
{
    //check to see if all windows in *a have SkipTaskbar & SkipPager set
    gboolean ret=FALSE;
    g_slist_foreach((GSList *)wnck_class_group_get_windows(a),(GFunc)checkit,&ret);
    return ret;
}

BDockTile * newanchor(gchar * image)
{
    BDockTile * bt=bdock_tile_new(BDOCK_CLASS_ANCHOR,NULL);
    GtkWidget * ch=gtk_image_new_from_pixbuf(gdk_pixbuf_new_from_file(image,NULL));
    bdock_tile_set_bg_from_file(bt,PIXMAPDIR "tile.tiff");
    bdock_tile_set_child(bt,ch);
    bdock_tile_place_child(bt);
    bdock_tile_register(bt);
    bdock_tile_show(bt);
    gtk_tooltips_set_tip(tips,GTK_WIDGET(bt->win),
            "Dock Anchor",NULL);
    return bt;
}

void newapp(WnckScreen * s, WnckClassGroup * a,BDockTile * anchor)
{
    BDockTile * bt=find_tile_by_class_name(
            wnck_class_group_get_res_class(a));
    if (bt)
    {
        GtkWidget * im = gtk_image_new_from_pixbuf(
                wnck_class_group_get_icon(a));
        bdock_tile_set_child(bt,im);
        bdock_tile_place_child(bt);
    }
    else
    {
        if (!confirmApp(a))
            return;
        bt=bdock_tile_new(BDOCK_CLASS_APP,anchor);
        bdock_tile_set_bg_from_file(bt,PIXMAPDIR "tile.tiff");
        GtkWidget * im = gtk_image_new_from_pixbuf(
                wnck_class_group_get_icon(a));
        bdock_tile_set_child(bt,im);
        bdock_tile_place_child(bt);
        bt->data.app.inactive_img=
            gdk_pixbuf_new_from_file(PIXMAPDIR "defaultAppIcon.xpm",NULL);
        bt->data.app.sticky=FALSE;
        bt->data.app.class_name=g_strdup(wnck_class_group_get_res_class(a));
        bt->data.app.class_group=a;
        bt->data.app.anchor=anchor;
        pid_t p = wnck_window_get_pid(
                wnck_class_group_get_windows(a)->data);
        if (p)
        {
            gchar * fp = g_strdup_printf("/proc/%d/cmdline",p);
            if (!g_file_get_contents(fp,&(bt->data.app.command),NULL,NULL))
            {
                bt->data.app.command=NULL;
            }
            g_free(fp);
        }
        bdock_tile_register(bt);
        if (bt->data.app.anchor->state.anchor==BDOCK_ANCHOR_STATE_COLLAPSED)
            bdock_tile_hide(bt);
        else
            bdock_tile_show(bt);
        gtk_tooltips_set_tip(tips,GTK_WIDGET(bt->win),
                wnck_class_group_get_res_class(a),NULL);
    }
}
void delswallow(GtkSocket * s, BDockTile * bt)
{
    bdock_tile_destroy(bt);
}

void app_gen_menu(BDockTile * bt, gpointer p)
{
    if (bt->class==BDOCK_CLASS_APP)
        bdock_tile_gen_menu(bt);
}

void clearpending(WnckScreen * s, WnckWindow * w,gpointer p)
{
    g_slist_foreach(tiles,(GFunc)app_gen_menu,NULL);
    pendingActivations=g_slist_remove_all(pendingActivations,w);
}

void set_icon_geometry_if(BDockTile * ba, WnckWindow * w)
{
    if (ba->class==BDOCK_CLASS_APP &&
            wnck_window_get_class_group(w)==ba->data.app.class_group)
    {
        do_set_icon_geometry(w,ba);
    }
}

void newswallow(WnckScreen * s, WnckWindow * w,BDockTile * anchor)
{
    BDockTile * bt=find_tile_by_class_group(wnck_window_get_class_group(w));
    if (bt)
        bdock_tile_gen_menu(bt);
    gdk_error_trap_push();
    XWMHints * h = XGetWMHints(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            wnck_window_get_xid(w));
    g_slist_foreach(tiles,(GFunc)set_icon_geometry_if,w);
    if ((h && h->initial_state==WithdrawnState) || has_dockapp_hint(w))
    {
        long su;
        GdkNativeWindow i = wnck_window_get_xid(w);
        bt=bdock_tile_new(BDOCK_CLASS_SWALLOW,anchor);
        bdock_tile_set_bg_from_file(bt,PIXMAPDIR "tile.tiff");
        set_dockapp_hint(w);
        GtkSocket * s = GTK_SOCKET(gtk_socket_new());
        XSizeHints * sh = XAllocSizeHints();
        int xp,yp,he,wi;
        gtk_container_set_border_width(GTK_CONTAINER(s),0);
        wnck_window_get_geometry(w,&xp,&yp,&wi,&he);
        if (wi==1 && he==1)
        {
            if (XGetWMNormalHints(
                        GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
                        i,sh,&su))
            {
                if (su & PSize)
                {
                    wi=sh->base_width;
                    he=sh->base_height;
                }
                else if (su & PMinSize)
                {
                    wi=sh->min_width;
                    he=sh->min_height;
                }
                else if (su & PMaxSize)
                {
                    wi=sh->max_width;
                    he=sh->max_height;
                }
                else if (su & USSize)
                {
                    wi=sh->width;
                    he=sh->height;
                }
            }
        }
        if (wi>64) wi=64;
        if (he>64) he=64;
        if (wi<=10) wi=64;
        if (he<=10) he=64;
        gtk_widget_set_size_request(GTK_WIDGET(s),wi,he);
        XFree(sh);
        bdock_tile_set_child(bt,GTK_WIDGET(s));
        gtk_socket_add_id(s,i);
        bdock_tile_place_child(bt);
        g_signal_connect(s,"plug-removed",G_CALLBACK(delswallow),bt);
        bt->data.swallow.id=i;
        bt->data.swallow.anchor=anchor;
        bdock_tile_register(bt);
        if (bt->data.app.anchor->state.anchor==BDOCK_ANCHOR_STATE_COLLAPSED)
            bdock_tile_hide(bt);
        else
            bdock_tile_show(bt);
        gtk_tooltips_set_tip(tips,GTK_WIDGET(bt->win),
                wnck_window_get_name(w),NULL);
    }
    if (h)
        XFree(h);
    XSync(GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),False);
    gdk_error_trap_pop();
}
void delapp(WnckScreen * screen, WnckClassGroup * group,gpointer p)
{
    BDockTile * tile=find_tile_by_class_group(group);
    if (!tile) return;
    if (!tile->data.app.sticky)
    {
        g_free(tile->data.app.class_name);
        bdock_tile_destroy(tile);
    }
    else
    {
        tile->data.app.class_group=NULL;
        bdock_tile_set_child(tile,
                gtk_image_new_from_pixbuf(
                    tile->data.app.inactive_img));
        bdock_tile_place_child(tile);
    }
}
int main(int argc,char ** argv)
{
    WnckScreen * s;
    BDockTile * a;
    gtk_init(&argc,&argv);
    moveCursor=gdk_cursor_new(GDK_FLEUR);
    tips=gtk_tooltips_new();
    gtk_tooltips_enable(tips);
    DockAppAtom=XInternAtom(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            "_BDOCK_DOCKAPP",False);
    RestackAtom=XInternAtom(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            "_NET_RESTACK_WINDOW",False);
    ActiveWinAtom=XInternAtom(
            GDK_DISPLAY_XDISPLAY(gdk_display_get_default()),
            "_NET_ACTIVE_WINDOW",False);
    width=gdk_screen_get_width(gdk_screen_get_default());
    height=gdk_screen_get_height(gdk_screen_get_default());
    s = wnck_screen_get(0);
    g_timeout_add(100,sync_timeout,NULL);
    a=newanchor(PIXMAPDIR "clip.tiff");
    g_signal_connect(s,"class_group_opened",G_CALLBACK(newapp),a);
    g_signal_connect(s,"window_opened",G_CALLBACK(newswallow),a);
    g_signal_connect(s,"window_closed",G_CALLBACK(clearpending),NULL);
    g_signal_connect(s,"class_group_closed",G_CALLBACK(delapp),NULL);
    g_signal_connect(s,"active_window_changed",
            G_CALLBACK(my_active_window_changed),NULL);
    gtk_main();
    return 0;
}
